/*
   Copyright (C) 2004 Michael Nelson Moran

   This file is part of OSCL.

   OSCL is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   OSCL is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with OSCL; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To negotiate other licensing arrangments/agreements contact
   the copyright holder:

   Michael N. Moran
   5009 Old Field Ct.         mike@mnmoran.org
   Kennesaw, GA, USA 30144    http://mnmoran.org
*/

#include "thread.h"
#include "oscl/error/fatal.h"
#include "oscl/error/info.h"
#include <pthread.h>
#include <unistd.h>
#include <stdio.h>

using namespace Oscl::Mt::Posix;

static void* entryPoint(void* runnable);

static volatile bool	keyCreated=false;
static pthread_key_t	key;
static pthread_attr_t	attributes;

Thread::Thread(	Oscl::Mt::Runnable&	runnable,
				const char*			name
				) throw():
		_runnable(runnable),
		_thread(),
		_syncSema(),
		_name(name)
		{
	printf("this=%8.8lX\n",(unsigned long)this);
	int err	= pthread_create(	&_thread,
								&attributes,
								&entryPoint,
								this
								);
	printf("that=%8.8lX\n",(unsigned long)this);
	if(err){
		Oscl::ErrorFatal::logAndExit("Thread::Thread() cannot create thread.\n");
		}
	}

Thread::~Thread(){
	int err	= pthread_cancel(_thread);
	if(err){
		Oscl::ErrorFatal::logAndExit("Thread::~Thread() cannot cancel thread.\n");
		}
	}

void	Thread::suspend() throw(){
	Oscl::ErrorFatal::logAndExit("Thread::suspend() not implemented.\n");
	}

void	Thread::suSuspend() throw(){
	Oscl::ErrorFatal::logAndExit("Thread::suSuspend() not implemented.\n");
	}

void	Thread::resume() throw(){
	Oscl::ErrorFatal::logAndExit("Thread::resume() not implemented.\n");
	}

void	Thread::suResume() throw(){
	Oscl::ErrorFatal::logAndExit("Thread::suResume() not implemented.\n");
	}

Oscl::Mt::Sema::SignalApi&	Thread::getSyncSignalApi() throw(){
	return _syncSema;
	}

void		Thread::syncWait() throw(){
	_syncSema.wait();
	}

const char* Thread::getName() throw(){
	return _name.getString();
	}

static void* entryPoint(void* vthread){
	Oscl::Error::Info::log("entryPoint()\n");
	Oscl::Mt::Posix::Thread*	thread	= (Thread*)vthread;
	Oscl::Mt::Thread*			thisThread	= thread;
	int err	= pthread_setspecific(key,thisThread);
	if(err){
		Oscl::ErrorFatal::logAndExit("entryPoint() cannot set key.\n");
		}
	unsigned long	id	= pthread_self();
	printf(	"thread = %8.8lX, _runnable = %8.8lX, pid=%lu, name = %s\n",
			(unsigned long)thread,
			(unsigned long)&thread->_runnable,
			id,
			thread->getName()
			);
	thread->_runnable.run();
	return 0;
	}

Oscl::Mt::Thread::Thread() throw(){
	if(!keyCreated){
		Oscl::ErrorFatal::logAndExit(	"Oscl::Mt::Posix::Thread::Thread() cannot create key.\n"
										"Oscl::Mt::Posix::Thread::initialize() must be invoked\n"
										"any of these threads are created.\n"
										);
		}
	}

Oscl::Mt::Thread::~Thread(){
	}

Oscl::Mt::Thread&	Oscl::Mt::Thread::getCurrent() throw() {
	if(!keyCreated){
		Oscl::ErrorFatal::logAndExit("Oscl::Mt::Thread::getCurrent(): call Thread::initialize() first.\n");
		}
	Oscl::Mt::Thread*	thread	= (Oscl::Mt::Thread*)pthread_getspecific(key);
	return *thread;
	}

void	Oscl::Mt::Thread::sleep(unsigned long tocks) throw(){
	// Assume tocks in micro-seconds
	struct timespec	ts;
	ts.tv_nsec	= (tocks%1000000)*1000;
	ts.tv_sec	= tocks/1000000;
	while(nanosleep(&ts,&ts));
	}

void Thread::initialize() throw(){
	int	err;
	if(keyCreated){
		return;
		}
	err	= pthread_attr_init(&attributes);
	size_t	default_stack_size;
	pthread_attr_getstacksize(&attributes,&default_stack_size);
	pthread_attr_setstacksize(&attributes,8*default_stack_size);
	printf("default stack size = %lu\n",(unsigned long)default_stack_size);
	err	= pthread_key_create(&key,(void(*)(void*))0);
	if(err){
		Oscl::ErrorFatal::logAndExit("Oscl::Mt::Posix::Thread::Thread() cannot create key.\n");
		}
	else{
		keyCreated=true;
		}
	}

